/* +++++++++++++++++++++++++++++ MAIL BOXES ++++++++++++++++++++++++++++++++ */

static inline void mbx_sem_signal(SEM *sem)
{
	unsigned long flags;
	RT_TASK *task;

	flags = rt_global_save_flags_and_cli();
	if ((task = (sem->queue.next)->task)) {
		sem->queue.next = task->queue.next;
		(task->queue.next)->prev = &(sem->queue);
		task->blocked_on.sem = NOTHING;
		if ((task->state &= ~(SEMAPHORE | DELAYED)) == READY) {
			rt_schedule();
		}
	} else {
		sem->count = 1;
	}
	rt_global_restore_flags(flags);
}

static inline void mbx_signal(MBX *mbx)
{
	unsigned long flags;
	RT_TASK *task;

	flags = rt_global_save_flags_and_cli();
	if ((task = mbx->waiting_task)) {
		task->blocked_on.task = mbx->waiting_task = NOTHING;
		if ((task->state &= ~(SUSPENDED | DELAYED)) == READY) {
			rt_schedule();
		}
	} else {
		mbx->mtx = 1;
	}
	rt_global_restore_flags(flags);
}

static inline int mbx_sem_wait(SEM *sem, RT_TASK *rt_current)
{
	unsigned long flags;
	QUEUE *q;

	flags = rt_global_save_flags_and_cli();
	if (!(sem->count)) {
		rt_current->state |= SEMAPHORE;
		rt_current->blocked_on.sem = sem;
		q = &(sem->queue);
#ifdef SEM_PRIORD
		while ((q = q->next) != &(sem->queue) &&
			 (q->task)->priority <= rt_current->priority);
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_schedule();
	} else {
		sem->count = 0;
	}
	rt_global_restore_flags(flags);
	return (int)(rt_current->blocked_on.sem);
}

static inline int mbx_wait(MBX *mbx, int *fravbs, RT_TASK *rt_current)
{
	unsigned long flags;

	flags = rt_global_save_flags_and_cli();
	if (!mbx->mtx || !(*fravbs)) {
		rt_current->state |= SUSPENDED;
		rt_current->blocked_on.task = SOMETHING;
		mbx->waiting_task = rt_current;
		rt_schedule();
	} else {
		mbx->mtx = 0;
	}
	rt_global_restore_flags(flags);
	return (int)(rt_current->blocked_on.task);
}

static inline int mbx_sem_wait_until(SEM *sem, RTIME time, RT_TASK *rt_current)
{
	int timed = 0;
	unsigned long flags;
	QUEUE *q;

	flags = rt_global_save_flags_and_cli();
	if (!(sem->count)) {
		rt_current->blocked_on.sem = sem;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->state |= (SEMAPHORE | DELAYED);
			q = &(sem->queue);
#ifdef SEM_PRIORD
			while ((q = q->next) != &(sem->queue) &&
				 (q->task)->priority <= rt_current->priority);
#endif
			rt_current->queue.prev = q->prev;
			rt_current->queue.next = q;
			(q->prev)->next = &(rt_current->queue);
			q->prev = &(rt_current->queue);
			rt_schedule();
		}
		if (rt_current->blocked_on.sem) {
			rt_current->blocked_on.sem = NOTHING;
			timed = 1;
		}
	} else {
		sem->count = 0;
	}
	rt_global_restore_flags(flags);
	return timed;
}

static inline int mbx_wait_until(MBX *mbx, int *fravbs, RTIME time, RT_TASK *rt_current)
{
	int timed = 0;
	unsigned long flags;

	flags = rt_global_save_flags_and_cli();
	if (!mbx->mtx || !(*fravbs)) {
		rt_current->blocked_on.task = SOMETHING;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->state |= DELAYED;
			mbx->waiting_task = rt_current;
			rt_schedule();
		}
		if (rt_current->blocked_on.task) {
			rt_current->blocked_on.task = mbx->waiting_task = NOTHING;
			timed = 1;
		}
	} else {
		mbx->mtx = 0;
	}
	rt_global_restore_flags(flags);
	return timed;
}

#define MOD_SIZE(indx) ((indx) < mbx->size ? (indx) : (indx) - mbx->size)

static inline int mbxput(MBX *mbx, char **msg, int msg_size)
{
	int tocpy, last_byte;

	while (mbx->frbs && msg_size > 0) {
		last_byte = MOD_SIZE(mbx->fbyte + mbx->avbs);
		if ((tocpy = mbx->size - last_byte) > msg_size) {
			tocpy = msg_size;
		}
		if (tocpy > mbx->frbs) {
			tocpy = mbx->frbs;
		}
		memcpy(mbx->bufadr + last_byte, *msg, tocpy);
		mbx->frbs -= tocpy;
		mbx->avbs += tocpy;
		msg_size  -= tocpy;
		*msg       += tocpy;
	}
	return msg_size;
}

static inline int mbxget(MBX *mbx, char **msg, int msg_size)
{
	int tocpy;

	while (mbx->avbs && msg_size > 0) {
		if ((tocpy = mbx->size - mbx->fbyte) > msg_size) {
			tocpy = msg_size;
		}
		if (tocpy > mbx->avbs) {
			tocpy = mbx->avbs;
		}
		memcpy(*msg, mbx->bufadr + mbx->fbyte, tocpy);
		mbx->fbyte = MOD_SIZE(mbx->fbyte + tocpy);
		mbx->frbs += tocpy;
		mbx->avbs -= tocpy;
		msg_size  -= tocpy;
		*msg       += tocpy;
	}
	return msg_size;
}

#define RT_MBX_MAGIC 0x3ad46e9b

#define CHK_MBX_MAGIC { if (mbx->magic != RT_MBX_MAGIC) { return -EINVAL; } }

int rt_mbx_init(MBX *mbx, int size)
{
	if (!(mbx->bufadr = kmalloc(size, GFP_KERNEL))) { 
		return -ENOMEM;
	}
	rt_sem_init(&(mbx->sndsem), 1);
	rt_sem_init(&(mbx->rcvsem), 1);
	mbx->magic = RT_MBX_MAGIC;
	mbx->size = mbx->frbs = size;
	mbx->waiting_task = 0;
	mbx->mtx = 1;
	mbx->fbyte = mbx->avbs = 0;
	return 0;
}


int rt_mbx_delete(MBX *mbx)
{
	CHK_MBX_MAGIC;
	mbx->magic = 0;
	if (rt_sem_delete(&(mbx->sndsem)) || rt_sem_delete(&(mbx->rcvsem))) {
		return -EFAULT;
	}
	kfree(mbx->bufadr); 
	return 0;
}


int rt_mbx_send(MBX *mbx, void *msg, int msg_size)
{
#ifdef __SMP__
	DECLARE_RT_CURRENT;
#endif

	CHK_MBX_MAGIC;
	if (mbx_sem_wait(&(mbx->sndsem), rt_current = rt_whoami())) {
		return msg_size;
	}
	while (msg_size) {
		if (mbx_wait(mbx, &mbx->frbs, rt_current)) {
			return msg_size;
		}
		msg_size = mbxput(mbx, (char **)(&msg), msg_size);
		mbx_signal(mbx);
	}
	mbx_sem_signal(&(mbx->sndsem));
	return 0;
}


int rt_mbx_send_wp(MBX *mbx, void *msg, int msg_size)
{
	unsigned long flags;

	CHK_MBX_MAGIC;
	flags = rt_global_save_flags_and_cli();
	if (mbx->sndsem.count && mbx->mtx && mbx->frbs) {
		mbx->sndsem.count = mbx->mtx = 0;
		rt_global_restore_flags(flags);
		mbxput(mbx, (char **)(&msg), msg_size > mbx->frbs ? mbx->frbs : msg_size);
		mbx_signal(mbx);
		mbx_sem_signal(&(mbx->sndsem));
		return 0;
	}
	rt_global_restore_flags(flags);
	return msg_size;
}


int rt_mbx_send_if(MBX *mbx, void *msg, int msg_size)
{
	unsigned long flags;

	CHK_MBX_MAGIC;
	flags = rt_global_save_flags_and_cli();
	if (mbx->sndsem.count && mbx->mtx && msg_size <= mbx->frbs) {
		mbx->sndsem.count = mbx->mtx = 0;
		rt_global_restore_flags(flags);
		mbxput(mbx, (char **)(&msg), msg_size);
		mbx_signal(mbx);
		mbx_sem_signal(&(mbx->sndsem));
		return 0;
	}
	rt_global_restore_flags(flags);
	return msg_size;
}


int rt_mbx_send_until(MBX *mbx, void *msg, int msg_size, RTIME time)
{
#ifdef __SMP__
	DECLARE_RT_CURRENT;
#endif

	CHK_MBX_MAGIC;
	if (mbx_sem_wait_until(&(mbx->sndsem), time, rt_current= rt_whoami())) {
		return msg_size;
	}
	while (msg_size) {
		if (mbx_wait_until(mbx, &mbx->frbs, time, rt_current)) {
			return msg_size;
		}
		msg_size = mbxput(mbx, (char **)(&msg), msg_size);
		mbx_signal(mbx);
	}
	mbx_sem_signal(&(mbx->sndsem));
	return 0;
}


int rt_mbx_send_timed(MBX *mbx, void *msg, int msg_size, RTIME delay)
{
	return rt_mbx_send_until(mbx, msg, msg_size, (oneshot_timer ? rdtsc(): rt_times.tick_time) + delay);
}


int rt_mbx_receive(MBX *mbx, void *msg, int msg_size)
{
#ifdef __SMP__
	DECLARE_RT_CURRENT;
#endif

	CHK_MBX_MAGIC;
	if (mbx_sem_wait(&(mbx->rcvsem), rt_current = rt_whoami())) {
		return msg_size;
	}
	while (msg_size) {
		if (mbx_wait(mbx, &mbx->avbs, rt_current)) {
			return msg_size;
		}
		msg_size = mbxget(mbx, (char **)(&msg), msg_size);
		mbx_signal(mbx);
	}
	mbx_sem_signal(&(mbx->rcvsem));
	return 0;
}


int rt_mbx_receive_wp(MBX *mbx, void *msg, int msg_size)
{
	unsigned long flags;

	CHK_MBX_MAGIC;
	flags = rt_global_save_flags_and_cli();
	if (mbx->rcvsem.count && mbx->mtx && mbx->avbs) {
		mbx->rcvsem.count = mbx->mtx = 0;
		rt_global_restore_flags(flags);
		mbxget(mbx, (char **)(&msg), msg_size > mbx->avbs ? mbx->avbs : msg_size);
		mbx_signal(mbx);
		mbx_sem_signal(&(mbx->rcvsem));
		return 0;
	}
	rt_global_restore_flags(flags);
	return msg_size;
}


int rt_mbx_receive_if(MBX *mbx, void *msg, int msg_size)
{
	unsigned long flags;

	CHK_MBX_MAGIC;
	flags = rt_global_save_flags_and_cli();
	if (mbx->rcvsem.count && mbx->mtx && msg_size <= mbx->avbs) {
		mbx->rcvsem.count = mbx->mtx = 0;
		rt_global_restore_flags(flags);
		mbxget(mbx, (char **)(&msg), msg_size);
		mbx_signal(mbx);
		mbx_sem_signal(&(mbx->rcvsem));
		return 0;
	}
	rt_global_restore_flags(flags);
	return msg_size;
}


int rt_mbx_receive_until(MBX *mbx, void *msg, int msg_size, RTIME time)
{
#ifdef __SMP__
	DECLARE_RT_CURRENT;
#endif

	CHK_MBX_MAGIC;
	if (mbx_sem_wait_until(&(mbx->rcvsem), time, rt_current= rt_whoami())) {
		return msg_size;
	}
	while (msg_size) {
		if (mbx_wait_until(mbx, &mbx->avbs, time, rt_current)) {
			return msg_size;
		}
		msg_size = mbxget(mbx, (char **)(&msg), msg_size);
		mbx_signal(mbx);
	}
	mbx_sem_signal(&(mbx->rcvsem));
	return 0;
}


int rt_mbx_receive_timed(MBX *mbx, void *msg, int msg_size, RTIME delay)
{
	return rt_mbx_receive_until(mbx, msg, msg_size, (oneshot_timer ? rdtsc(): rt_times.tick_time) + delay);
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
